/*
 * tronMove.c
 * Code related to movement of objects for Homework 4.
 */

#include "GP142.h"
#include "tron.h"
#include <math.h>

/******************************************************************************
 * Symbolic constants.  There are other symbolic constants in tron.h.
 ******************************************************************************/

/* Speeds */
#define PLAYER_SPEED	10
#define BULLET_SPEED	30
#define ENEMY_SPEED		4

/* This constant is used in determining whether a bullet has hit an enemy.
 * It determines how small of pieces the bullet's path is broken up into
 * when doing hit detection.
 * It is *VERY IMPORANT* that this constant be a factor of BULLET_SPEED!
 * If it isn't, the code *will* go into an infinite loop!
 */
#define HIT_DETECTION_STEP_SIZE	5

/******************************************************************************
 * Function prototypes.  See the function definitions for 
 * descriptions of what they do.  Other prototypes are in tron.h.
 ******************************************************************************/
int directionToPlayer(int x, int y, int playerX, int playerY);

/******************************************************************************
 * moveObject
 *
 * Move an object in a given direction by a given amount.  This function
 * should be used for ALL movement behavior in the game, since it all
 * basically comes down to the same idea.  This is preferable to rewriting
 * the movement code for each type of object on the screen.
 *
 * Parameters:	x, y: pointers to the object's coordinates.
 *				width, height: object's dimensions
 *				direction: the direction to move the object in.
 *				speed: the distance to move the object, in pixels.
 */
void moveObject(int *x, int *y, int width, int height, int direction, int speed)
{
	/* If the object isn't stopped, update its position. */
	if (direction > 0)
	{
		/* xComponent and yComponent return -1, 0, or 1 to tell you the sign
		 * of the change to position in each direction. */
		*x += speed * xComponent(direction);
		*y += speed * yComponent(direction);
	}
	
	/* Prevent objects from going off the edge of the screen. */
	if (*x > GP142_XMAX - BOUNDARY_WIDTH - width)
	{
		*x = GP142_XMAX - BOUNDARY_WIDTH - width;
	} 
	else if (*x < -GP142_XMAX + BOUNDARY_WIDTH) 
	{
		*x = -GP142_XMAX + BOUNDARY_WIDTH;
	}

	if (*y > GP142_YMAX - BANNER_HEIGHT - height)
	{
		*y = GP142_YMAX - BANNER_HEIGHT - height;
	}
	else if (*y < -GP142_YMAX + BOUNDARY_WIDTH) 
	{
		*y = -GP142_YMAX + BOUNDARY_WIDTH;
	}
}

/******************************************************************************
 * movePlayer
 *
 * Move the player in the indicated direction.
 *
 * Parameters: pointers to the player's current coordinates, as well as
 *				the direction to move in.
 */
void movePlayer(int *x, int *y, int playerFacing)
{
	moveObject(x, y, PLAYER_WIDTH, PLAYER_HEIGHT, playerFacing, PLAYER_SPEED);
}

/******************************************************************************
 * moveEnemy
 *
 * Move the enemy.  No direction is needed because the enemy decides where to
 * go by itself.  It does need to know the player's coordinates, though.
 *
 * Parameters: x and y: pointers to the enemy's current position
 *			   playerX and playerY: the player's position
 *			   intelligence: a number from 0.0 to 1.0 indicating how smart the
 *					enemy is (higher is smarter).
 */
void moveEnemy(int *x, int *y, int playerX, int playerY)
{
	int enemyCenterX = *x + ENEMY_WIDTH/2,
		enemyCenterY = *y + ENEMY_HEIGHT/2,
		playerCenterX = playerX + PLAYER_WIDTH/2,
		playerCenterY = playerY + PLAYER_HEIGHT/2;
	int direction = directionToPlayer(enemyCenterX, enemyCenterY,
									  playerCenterX, playerCenterY);
	moveObject(x, y, ENEMY_WIDTH, ENEMY_HEIGHT, direction, ENEMY_SPEED);
}

/******************************************************************************
 * moveBullet
 *
 * Move the bullet in the indicated direction.
 *
 * Parameters: pointers to the bullet's current coordinates, as well as
 *				the direction to move it in.
 */
void moveBullet(int *x, int *y, int *bulletStatus, int bulletFacing)
{
	moveObject(x, y, BULLET_WIDTH, BULLET_HEIGHT, bulletFacing, BULLET_SPEED);

	/* Detect the bullet hitting the edge of the screen, and kill it if it 
	 * does so. */
	if (*x >= GP142_XMAX - BOUNDARY_WIDTH - BULLET_WIDTH
		|| *x <= -GP142_XMAX + BOUNDARY_WIDTH + BULLET_WIDTH
		|| *y >= GP142_YMAX - BOUNDARY_WIDTH - BANNER_HEIGHT - BULLET_HEIGHT
		|| *y <= -GP142_YMAX + BOUNDARY_WIDTH + BULLET_HEIGHT) 
	{
		*bulletStatus = BULLET_DEAD;
	}

}

/******************************************************************************
 * rectanglesOverlap
 *
 * Determine whether two rectangular regions on the screen overlap.  This
 * is how we tell whether two objects have collided.
 * This has all sorts of uses, and will be called a LOT!
 *
 * Parameters: x1, y1: coordinates of rectangle 1's top left corner
 *			   w1, h1: rectangle 1's height and width
 *			   x2, y2, w2, h2: as above, but for rectangle 2
 */
int rectanglesOverlap(int x1, int y1, int w1, int h1,
					  int x2, int y2, int w2, int h2)
{
	/* This is a big nasty conditional statement.  It tells you whether two 
	 * rectangles overlap.  Yes, it's ugly.  No, you shouldn't write code like
	 * this.  Just take a deep breath, pretend it all makes sense, and pretend
	 * you never saw it. */
	if( (y1<=y2&&y1+h1>=y2 || y1<=y2+h2&&y1+h1>=y2+h2 || y1>=y2&&y1+h1<=y2+h2) &&
	    (x1<=x2&&x1+w1>=x2 || x1<=x2+w2&&x1+w1>=x2+w2 || x1>=x2&&x1+w1<=x2+w2) )
	{
            return TRUE;
	}
	return FALSE;
}

/******************************************************************************
 * rectangleHitByBullet
 *
 * Determine whether a rectangle has been hit by a shot.  To determine this,
 * we need the new and previous positions of the bullet.
 *
 * Parameters:	rectX, rectY: coordinates of rectangle's top left corner
 *				width, height: rectangle's dimensions
 *				bulletX1, bulletY1: coordinates of bullet's old position
 *				bulletX2, bulletY2: coordinates of bullet's new position
 */
int rectangleHitByBullet(int rectX, int rectY, int width, int height,
						 int bulletX1, int bulletY1, int bulletX2, int bulletY2,
						 int bulletFacing)
{
	/* We can't just do hit detection based on rectangle overlap with the
	 * new bullet position, because this might allow the bullet to "hop"
	 * over the enemy.
	 * Instead, we check the bullet several times along its path to see
	 * if it ever intersects the enemy. */
	while (bulletX1 != bulletX2 || bulletY1 != bulletY2)
	{
		if (rectanglesOverlap(rectX, rectY, width, height,
							  bulletX1, bulletY1, BULLET_WIDTH, BULLET_HEIGHT))
		{
			return TRUE;
		}

		/* Update the X1/Y1 coordinates to check the next position on the path.
		 * Note that this has no effect on the position of the actual bullet!
		 * We're just using moveObject to find out where the actual bullet 
		 * *would* be if it moved to the next step along the path.  This is a
		 * sort of "simulation" of the actual bullet's movement, accomplished 
		 * by using variables containing *copies* of the actual bullet's
		 * coordinates. */
		moveObject(&bulletX1, &bulletY1, BULLET_WIDTH, BULLET_HEIGHT,
			bulletFacing, HIT_DETECTION_STEP_SIZE);
	}

	return FALSE;
}

/******************************************************************************
 * directionToPlayer
 *
 * Find the best direction to travel to move towards the player.
 *
 * The details of how this is done aren't particularly important, and are in
 * fact a little bit messy.  It would be cool to come up with a more elegant
 * solution.
 *
 * Basically, the playing field is divided into 8 regions, each comprising a
 * 22.5 degree arc surrounding one of the 8 directions.  This code determines
 * which region the player is in.
 */
int directionToPlayer(int x, int y, int playerX, int playerY)
{
	double dx = playerX - x;
	double dy = playerY - y;

	/* fabs is like abs, except it's used for doubles. */
	if (dy >= 0.0 && fabs(dy/dx) > 0.25)
	{
		if (fabs(dy/dx) > 0.75)
		{
			return UP;
		}
		else if (dx < 0.0) 
		{
			return UP_LEFT;
		}
		return UP_RIGHT;
	}
	else if (dy < 0.0 && fabs(dy/dx) > 0.25) 
	{
		if (fabs(dy/dx) > 0.75)
		{
			return DOWN;
		}
		else if (dx < 0.0) 
		{
			return DOWN_LEFT;
		}
		return DOWN_RIGHT;
	}
	if (dx < 0)
	{
		return LEFT;
	}
	return RIGHT;
}

/******************************************************************************
 * xComponent
 *
 * Given a direction (one of the symbolic constants defined), return:
 *		-1 if the direction points to the left
 *		0  if the direction doesn't point right or left
 *		1  if the direction points to the right
 *		2  if the supplied value is not a valid direction (error code)
 *
 * If this is unclear, look at the implementation for examples.
 */
int xComponent(int direction)
{
	/* Note that we take the absolute value, to deal with objects that are
	 * stopped. */
	switch (abs(direction))
	{
	/* Note that there are no break statements in this switch block.  That
	 * means that cases will "fall through" to the next one automatically.
	 * However, they don't go all the way to the bottom, because once you
	 * hit a return statement the function returns to its caller -- it
	 * doesn't keep going through the rest of the switch statement. */
	case UP_LEFT:
	case LEFT:
	case DOWN_LEFT:
		return -1;
	case UP:
	case DOWN:
		return 0;
	case UP_RIGHT:
	case RIGHT:
	case DOWN_RIGHT:
		return 1;
	default:
		return 2;
	}
}

/******************************************************************************
 * yComponent
 *
 * Given a direction (one of the symbolic constants defined), return:
 *		-1 if the direction points up
 *		0  if the direction doesn't point up or down
 *		1  if the direction points down
 *		2  if the supplied value is not a valid direction (error code)
 *
 * If this is unclear, look at the implementation for examples.
 */
int yComponent(int direction)
{
	/* Note that we take the absolute value, to deal with objects that are
	 * stopped. */
	switch (abs(direction))
	{
	/* Note that there are no break statements in this switch block.  That
	 * means that cases will "fall through" to the next one automatically.
	 * However, they don't go all the way to the bottom, because once you
	 * hit a return statement the function returns to its caller -- it
	 * doesn't keep going through the rest of the switch statement. */
	case UP_LEFT:
	case UP:
	case UP_RIGHT:
		return 1;
	case LEFT:
	case RIGHT:
		return 0;
	case DOWN_LEFT:
	case DOWN:
	case DOWN_RIGHT:
		return -1;
	default:
		return 2;
	}
}
